@@ -3,23 +3,17 @@
  *
  *	TclXML driver for James Clark's expat XML parser
  *
- * Copyright (c) 1998-2000 Steve Ball, Zveno Pty Ltd
+ * Copyright (c) 1998-2003 Steve Ball, Zveno Pty Ltd
  *
- * Zveno Pty Ltd makes this software and associated documentation
- * available free of charge for any purpose.  You may make copies
- * of the software but you must include all of this notice on any copy.
+ * See the file "LICENSE" for information on usage and
+ * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * Zveno Pty Ltd does not warrant that this software is error free
- * or fit for any purpose.  Zveno Pty Ltd disclaims any liability for
- * all claims, expenses, losses, damages and costs any user may incur
- * as a result of using, copying or modifying the software.
- *
- * $Id: tclexpat.c,v 1.14 2002/12/10 03:29:27 balls Exp $
+ * $Id: tclexpat.c,v 1.19 2004/02/20 09:15:39 balls Exp $
  *
  */
 
 #include <tcl.h>
-#include "tclxml.h"
+#include <tclxml/tclxml.h>
 #include <xmlparse.h>
 
 #undef TCL_STORAGE_CLASS
@@ -49,7 +43,8 @@ static int	TclExpatDelete _ANSI_ARGS_((ClientData clientData));
 static int	TclExpatParse _ANSI_ARGS_((ClientData clientData, 
                     char *data, int len, int final));
 static int	TclExpatConfigure _ANSI_ARGS_((ClientData clientdata, 
-                    int objc, Tcl_Obj *CONST objv[]));
+					       Tcl_Obj *CONST optionPtr,
+					       Tcl_Obj *CONST valuePtr));
 static int	TclExpatCget _ANSI_ARGS_((ClientData clientData, 
                     int objc, Tcl_Obj *CONST objv[]));
 static int	TclExpatGet _ANSI_ARGS_((ClientData clientData, 
@@ -179,7 +174,7 @@ Tclexpat_Init (interp)
     return TCL_ERROR;
   }
 
-  if (Tcl_PkgProvide(interp, "xml::expat", "2.5") != TCL_OK) {
+  if (Tcl_PkgProvide(interp, "xml::expat", TCLXML_VERSION) != TCL_OK) {
     return TCL_ERROR;
   }
 
@@ -411,12 +406,13 @@ TclExpatParse(clientData, data, len, final)
  */
 
 static int
-TclExpatConfigure (clientData, objc, objv)
+TclExpatConfigure (clientData, optionPtr, valuePtr)
      ClientData clientData;
-     int objc;
-     Tcl_Obj *CONST objv[];
+     Tcl_Obj *CONST optionPtr;
+     Tcl_Obj *CONST valuePtr;
 {
   TclExpatInfo *expat = (TclExpatInfo *) clientData;
+  int index, value, bool;
 
   static CONST84 char *switches[] = {
     "-baseurl", 
@@ -440,67 +436,58 @@ TclExpatConfigure (clientData, objc, objv)
       EXPAT_PARAMENTITYPARSINGNOTSTANDALONE
   };
 
-  int index, value, bool;
-  Tcl_Obj *CONST *objPtr = objv;
+  if (Tcl_GetIndexFromObj(expat->interp, optionPtr, switches,
+			  "switch", 0, &index) != TCL_OK) {
+    Tcl_ResetResult(expat->interp);
+    return TCL_OK;
+  }
+  switch ((enum switches) index) {
 
-  while (objc > 1) {
-    if (Tcl_GetIndexFromObj(expat->interp, objPtr[0], switches,
-			    "switch", 0, &index) != TCL_OK) {
-      Tcl_ResetResult(expat->interp);
-      return TCL_OK;
+  case EXPAT_BASE:			/* -base */
+
+    if (XML_SetBase(expat->parser, (const XML_Char*)Tcl_GetString(valuePtr)) == 0) {
+      Tcl_SetResult(expat->interp, "unable to set base URL", NULL);
+      return TCL_ERROR;
     }
-    switch ((enum switches) index) {
+    break;
 
-      case EXPAT_BASE:			/* -base */
+  case EXPAT_DEFAULTEXPANDINTERNALENTITIES: 
 
-	if (XML_SetBase(expat->parser, (const XML_Char*)Tcl_GetString(objPtr[1])) == 0) {
-	  Tcl_SetResult(expat->interp, "unable to set base URL", NULL);
-	  return TCL_ERROR;
-	}
-	break;
-
-      case EXPAT_DEFAULTEXPANDINTERNALENTITIES: 
-
-        /* -defaultexpandinternalentities */
-	/* ericm@scriptics */
-	if (Tcl_GetBooleanFromObj(expat->interp, objPtr[1], &bool) != TCL_OK) {
-	  return TCL_ERROR;
-	}
+    /* -defaultexpandinternalentities */
+    /* ericm@scriptics */
+    if (Tcl_GetBooleanFromObj(expat->interp, valuePtr, &bool) != TCL_OK) {
+      return TCL_ERROR;
+    }
 	
-        XML_SetDefaultExpandInternalEntities(expat->parser, bool);
+    XML_SetDefaultExpandInternalEntities(expat->parser, bool);
 
-	break;
-
-      case EXPAT_PARAMENTITYPARSING: /* -paramentityparsing */
-	  /* ericm@scriptics */
-	  if (Tcl_GetIndexFromObj(expat->interp, objPtr[1], paramEntityParsingValues,
-		  "value", 0, &value) != TCL_OK) {
-	      return TCL_ERROR;
-	  }
-	  switch ((enum paramEntityParsingValues) value) {
-	      case EXPAT_PARAMENTITYPARSINGALWAYS:
-		  XML_SetParamEntityParsing(expat->parser,
-			  XML_PARAM_ENTITY_PARSING_ALWAYS);
-		  break;
-	      case EXPAT_PARAMENTITYPARSINGNEVER:
-		  XML_SetParamEntityParsing(expat->parser,
-			  XML_PARAM_ENTITY_PARSING_NEVER);
-		  break;
-	      case EXPAT_PARAMENTITYPARSINGNOTSTANDALONE:
-		  XML_SetParamEntityParsing(expat->parser,
-			  XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE);
-		  break;
-	  }
-	  break;
-
-    default:
-      return TCL_OK;
-	  break;
+    break;
 
+  case EXPAT_PARAMENTITYPARSING: /* -paramentityparsing */
+    /* ericm@scriptics */
+    if (Tcl_GetIndexFromObj(expat->interp, valuePtr, paramEntityParsingValues,
+			    "value", 0, &value) != TCL_OK) {
+      return TCL_ERROR;
     }
+    switch ((enum paramEntityParsingValues) value) {
+    case EXPAT_PARAMENTITYPARSINGALWAYS:
+      XML_SetParamEntityParsing(expat->parser,
+				XML_PARAM_ENTITY_PARSING_ALWAYS);
+      break;
+    case EXPAT_PARAMENTITYPARSINGNEVER:
+      XML_SetParamEntityParsing(expat->parser,
+				XML_PARAM_ENTITY_PARSING_NEVER);
+      break;
+    case EXPAT_PARAMENTITYPARSINGNOTSTANDALONE:
+      XML_SetParamEntityParsing(expat->parser,
+				XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE);
+      break;
+    }
+    break;
 
-    objPtr += 2;
-    objc -= 2;
+  default:
+    return TCL_OK;
+    break;
 
   }
 
@@ -625,19 +612,29 @@ static void TclExpatElementStartHandler(userdata, name, atts)
      const XML_Char **atts;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
-  Tcl_Obj *attListObj;
+  Tcl_Obj *attListObj, *nameObj;
   int count;
 
   attListObj = Tcl_NewListObj(0, NULL);
+  Tcl_IncrRefCount(attListObj);
   for (count = 0; atts[count]; count += 2) {
-    Tcl_ListObjAppendElement(NULL, attListObj, Tcl_NewStringObj((XML_Char*)atts[count], -1));
-    Tcl_ListObjAppendElement(NULL, attListObj, Tcl_NewStringObj((XML_Char*)atts[count + 1], -1));
+    Tcl_ListObjAppendElement(NULL, attListObj,
+			     Tcl_NewStringObj((XML_Char*)atts[count], -1));
+    Tcl_ListObjAppendElement(NULL, attListObj,
+			     Tcl_NewStringObj((XML_Char*)atts[count + 1], -1));
   }
 
+  nameObj = Tcl_NewStringObj((XML_Char *)name, -1);
+  Tcl_IncrRefCount(nameObj);
   TclXML_ElementStartHandler(
         expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)name, -1),
-	attListObj);
+	nameObj,
+	NULL,
+	attListObj,
+	NULL);
+
+  Tcl_DecrRefCount(nameObj);
+  Tcl_DecrRefCount(attListObj);
 }
 
 static void TclExpatElementEndHandler(userdata, name)
@@ -645,10 +642,12 @@ static void TclExpatElementEndHandler(userdata, name)
      const XML_Char *name;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *nameObj = Tcl_NewStringObj((XML_Char*)name, -1);
+  Tcl_IncrRefCount(nameObj);
 
-  TclXML_ElementEndHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)name, -1));
+  TclXML_ElementEndHandler(expat->xmlinfo, nameObj);
+
+  Tcl_DecrRefCount(nameObj);
 }
 
 static void TclExpatCharacterDataHandler(userdata, s, len)
@@ -657,10 +656,12 @@ static void TclExpatCharacterDataHandler(userdata, s, len)
      int len;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *dataObj = Tcl_NewStringObj((XML_Char*)s, len);
+  Tcl_IncrRefCount(dataObj);
 
-  TclXML_CharacterDataHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)s, len));
+  TclXML_CharacterDataHandler(expat->xmlinfo, dataObj);
+
+  Tcl_DecrRefCount(dataObj);
 }
 
 static void TclExpatProcessingInstructionHandler(userdata, target, data)
@@ -669,11 +670,15 @@ static void TclExpatProcessingInstructionHandler(userdata, target, data)
      const XML_Char *data;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *dataObj = Tcl_NewStringObj((XML_Char*)data, -1);
+  Tcl_Obj *targetObj = Tcl_NewStringObj((XML_Char*)target, -1);
+  Tcl_IncrRefCount(dataObj);
+  Tcl_IncrRefCount(targetObj);
 
-  TclXML_ProcessingInstructionHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)target, -1),
-	Tcl_NewStringObj((XML_Char*)data, -1));
+  TclXML_ProcessingInstructionHandler(expat->xmlinfo, targetObj, dataObj);
+
+  Tcl_DecrRefCount(dataObj);
+  Tcl_DecrRefCount(targetObj);
 }
 
 static void TclExpatDefaultHandler(userdata, s, len)
@@ -682,10 +687,12 @@ static void TclExpatDefaultHandler(userdata, s, len)
      int len;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *dataObj = Tcl_NewStringObj((XML_Char*)s, len);
+  Tcl_IncrRefCount(dataObj);
 
-  TclXML_DefaultHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)s, len));
+  TclXML_DefaultHandler(expat->xmlinfo, dataObj);
+
+  Tcl_DecrRefCount(dataObj);
 }
 
 static void TclExpatUnparsedDeclHandler(userdata, entityName, base, systemId, publicId, notationName)
@@ -697,15 +704,28 @@ static void TclExpatUnparsedDeclHandler(userdata, entityName, base, systemId, pu
      const XML_Char *notationName;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *entityObj, *baseObj, *systemIdObj, *publicIdObj, *notationNameObj;
 
-  TclXML_UnparsedDeclHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)entityName, -1),
-	Tcl_NewStringObj((XML_Char*)base, -1),
-	Tcl_NewStringObj((XML_Char*)systemId, -1),
-	Tcl_NewStringObj((XML_Char*)publicId, -1),
-	Tcl_NewStringObj((XML_Char*)notationName, -1)
-	);
+  entityObj = Tcl_NewStringObj((XML_Char*)entityName, -1);
+  baseObj = Tcl_NewStringObj((XML_Char*)base, -1);
+  systemIdObj = Tcl_NewStringObj((XML_Char*)systemId, -1);
+  publicIdObj = Tcl_NewStringObj((XML_Char*)publicId, -1);
+  notationNameObj = Tcl_NewStringObj((XML_Char*)notationName, -1);
+  
+  Tcl_IncrRefCount(entityObj);
+  Tcl_IncrRefCount(baseObj);
+  Tcl_IncrRefCount(systemIdObj);
+  Tcl_IncrRefCount(publicIdObj);
+  Tcl_IncrRefCount(notationNameObj);
+
+  TclXML_UnparsedDeclHandler(expat->xmlinfo, 
+          entityObj, baseObj, systemIdObj, publicIdObj, notationNameObj);
+
+  Tcl_DecrRefCount(entityObj);
+  Tcl_DecrRefCount(baseObj);
+  Tcl_DecrRefCount(systemIdObj);
+  Tcl_DecrRefCount(publicIdObj);
+  Tcl_DecrRefCount(notationNameObj);
 }
 
 static void TclExpatNotationDeclHandler(userdata, notationName, base, systemId, publicId)
@@ -716,14 +736,25 @@ static void TclExpatNotationDeclHandler(userdata, notationName, base, systemId,
      const XML_Char *publicId;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *notationNameObj, *baseObj, *systemIdObj, *publicIdObj;
 
-  TclXML_NotationDeclHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)notationName, -1),
-	Tcl_NewStringObj((XML_Char*)base, -1),
-	Tcl_NewStringObj((XML_Char*)systemId, -1),
-	Tcl_NewStringObj((XML_Char*)publicId, -1)
-	);
+  notationNameObj = Tcl_NewStringObj((XML_Char*)notationName, -1);
+  baseObj = Tcl_NewStringObj((XML_Char*)base, -1);
+  systemIdObj = Tcl_NewStringObj((XML_Char*)systemId, -1);
+  publicIdObj = Tcl_NewStringObj((XML_Char*)publicId, -1);
+
+  Tcl_IncrRefCount(notationNameObj);
+  Tcl_IncrRefCount(baseObj);
+  Tcl_IncrRefCount(systemIdObj);
+  Tcl_IncrRefCount(publicIdObj);
+
+  TclXML_NotationDeclHandler(expat->xmlinfo, 
+          notationNameObj, baseObj, systemIdObj, publicIdObj);
+
+  Tcl_DecrRefCount(notationNameObj);
+  Tcl_DecrRefCount(baseObj);
+  Tcl_DecrRefCount(systemIdObj);
+  Tcl_DecrRefCount(publicIdObj);
 }
 
 static void TclExpatCommentHandler(userdata, data)
@@ -731,10 +762,12 @@ static void TclExpatCommentHandler(userdata, data)
      const XML_Char *data;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *dataObj = Tcl_NewStringObj((XML_Char*)data, -1);
+  Tcl_IncrRefCount(dataObj);
 
-  TclXML_CommentHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)data, -1));
+  TclXML_CommentHandler(expat->xmlinfo, dataObj);
+
+  Tcl_DecrRefCount(dataObj);
 }
 
 static int TclExpatNotStandaloneHandler(userdata)
@@ -791,13 +824,25 @@ TclExpatExternalEntityRefHandler(parser, context, base, systemId, publicId)
      const XML_Char *publicId;
 {
   TclExpatInfo *expat = (TclExpatInfo *) XML_GetUserData(parser);
+  Tcl_Obj *contextObj, *baseObj, *systemIdObj, *publicIdObj;
 
-  TclXML_ExternalEntityRefHandler(
-        (ClientData) expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)context, -1), 
-	Tcl_NewStringObj((XML_Char*)base, -1), 
-	Tcl_NewStringObj((XML_Char*)systemId, -1), 
-	Tcl_NewStringObj((XML_Char*)publicId, -1));
+  contextObj = Tcl_NewStringObj((XML_Char*)context, -1);
+  baseObj = Tcl_NewStringObj((XML_Char*)base, -1);
+  systemIdObj = Tcl_NewStringObj((XML_Char*)systemId, -1);
+  publicIdObj = Tcl_NewStringObj((XML_Char*)publicId, -1);
+
+  Tcl_IncrRefCount(contextObj);
+  Tcl_IncrRefCount(baseObj);
+  Tcl_IncrRefCount(systemIdObj);
+  Tcl_IncrRefCount(publicIdObj);
+
+  TclXML_ExternalEntityRefHandler((ClientData) expat->xmlinfo, 
+          contextObj, baseObj, systemIdObj, publicIdObj);
+
+  Tcl_DecrRefCount(contextObj);
+  Tcl_DecrRefCount(baseObj);
+  Tcl_DecrRefCount(systemIdObj);
+  Tcl_DecrRefCount(publicIdObj);
 
   return 1;
 }
@@ -826,12 +871,16 @@ TclExpatUnknownEncodingHandler(encodingHandlerData, name, info)
      XML_Encoding *info;
 {
   int result;
+  Tcl_Obj *nameObj = Tcl_NewStringObj((XML_Char*)name, -1);
+  Tcl_IncrRefCount(nameObj);
 
   result = TclXML_UnknownEncodingHandler(
                  encodingHandlerData, 
-		 Tcl_NewStringObj((XML_Char*)name, -1), 
+		 nameObj,
 		 (ClientData) info);
 
+  Tcl_DecrRefCount(nameObj);
+
   return result;
 }
 
@@ -841,18 +890,24 @@ static void TclExpatElementDeclHandler(userdata, name, contentspec)
      XML_Char ***contentspec;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
-  Tcl_Obj *contentspecObj;
+  Tcl_Obj *contentspecObj, *nameObj;
   int count;
 
   contentspecObj = Tcl_NewListObj(0, NULL);
+  Tcl_IncrRefCount(contentspecObj);
+
   for (count = 0; contentspec[0][count] != NULL; count++) {
-    Tcl_ListObjAppendElement(expat->interp, contentspecObj, Tcl_NewStringObj(contentspec[0][count], -1));
+    Tcl_ListObjAppendElement(expat->interp, contentspecObj, 
+            Tcl_NewStringObj(contentspec[0][count], -1));
   }
 
-  TclXML_ElementDeclHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)name, -1),
-	contentspecObj);
+  nameObj = Tcl_NewStringObj((XML_Char*)name, -1);
+  Tcl_IncrRefCount(nameObj);
+
+  TclXML_ElementDeclHandler(expat->xmlinfo, nameObj, contentspecObj);
+
+  Tcl_DecrRefCount(nameObj);
+  Tcl_DecrRefCount(contentspecObj);
 }
 
 static void TclExpatAttlistDeclHandler(userdata, name, attributes)
@@ -861,18 +916,26 @@ static void TclExpatAttlistDeclHandler(userdata, name, attributes)
      XML_Char ***attributes;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
-  Tcl_Obj *attdefnObj;
+  Tcl_Obj *attdefnObj, *nameObj;
   int count;
 
   attdefnObj = Tcl_NewListObj(0, NULL);
+  Tcl_IncrRefCount(attdefnObj);
+
   for (count = 0; (*attributes)[count] != NULL; count++) {
-    Tcl_ListObjAppendElement(expat->interp, attdefnObj, Tcl_NewStringObj(*attributes[count], -1));
+    Tcl_ListObjAppendElement(expat->interp, attdefnObj,
+			     Tcl_NewStringObj((*attributes)[count], -1));
   }
 
+  nameObj = Tcl_NewStringObj((XML_Char*)name, -1);
+  Tcl_IncrRefCount(nameObj);
   TclXML_AttlistDeclHandler(
         expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)name, -1),
+	nameObj,
 	attdefnObj);
+
+  Tcl_DecrRefCount(nameObj);
+  Tcl_DecrRefCount(attdefnObj);
 }
 
 static void TclExpatStartDoctypeDeclHandler(userdata, name)
@@ -880,10 +943,12 @@ static void TclExpatStartDoctypeDeclHandler(userdata, name)
      const XML_Char *name;
 {
   TclExpatInfo *expat = (TclExpatInfo *) userdata;
+  Tcl_Obj *nameObj = Tcl_NewStringObj((XML_Char*)name, -1);
+  Tcl_IncrRefCount(nameObj);
 
-  TclXML_StartDoctypeDeclHandler(
-        expat->xmlinfo,
-	Tcl_NewStringObj((XML_Char*)name, -1));
+  TclXML_StartDoctypeDeclHandler(expat->xmlinfo, nameObj);
+
+  Tcl_DecrRefCount(nameObj);
 }
 
 static void TclExpatEndDoctypeDeclHandler(userdata)
