@@ -3,52 +3,51 @@
  *
  *  Generic interface to XML parsers.
  *
- * Copyright (c) 1998-2003 Steve Ball, Zveno Pty Ltd
+ * Copyright (c) 1998-2004 Steve Ball, Zveno Pty Ltd
  *
- * Zveno Pty Ltd makes this software and associated documentation
- * available free of charge for any purpose.  You may make copies
- * of the software but you must include all of this notice on any copy.
+ * See the file "LICENSE" for information on usage and
+ * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * Zveno Pty Ltd does not warrant that this software is error free
- * or fit for any purpose.  Zveno Pty Ltd disclaims any liability for
- * all claims, expenses, losses, damages and costs any user may incur
- * as a result of using, copying or modifying the software.
- *
- * $Id: tclxml.c,v 1.25 2003/02/25 04:09:00 balls Exp $
+ * $Id: tclxml.c,v 1.29 2004/02/26 05:12:21 balls Exp $
  *
  */
 
-#include "tclxml.h"
+#include <tclxml/tclxml.h>
+#include <string.h>
 
 #define TCL_DOES_STUBS \
     (TCL_MAJOR_VERSION > 8 || TCL_MAJOR_VERSION == 8 && (TCL_MINOR_VERSION > 1 || \
     (TCL_MINOR_VERSION == 1 && TCL_RELEASE_LEVEL == TCL_FINAL_RELEASE)))
+#ifndef TCLXML_DOES_STUBS
+# define TCLXML_DOES_STUBS TCL_DOES_STUBS
+#endif /* TCL_DOES_STUBS */
 
 /*
  * The structure below is used to manage package options.
  */
 
-typedef struct TclXMLConfigInfo {
+typedef struct ThreadSpecificData {
   TclXML_ParserClassInfo *defaultParser;    /* Current default parser */
-} TclXMLConfigInfo;
+  Tcl_HashTable *registeredParsers;         /* All known parser classes */
 
-static TclXMLConfigInfo *xmlconfig;
-static Tcl_HashTable registeredParsers;
+  /*
+   * Retain a pointer to the whitespace variable
+   */
 
-/*
- * Retain a pointer to the whitespace variable
- */
+  Tcl_Obj *whitespaceRE;
+
+  /*
+   * Counter to generate unique command names
+   */
+
+  int uniqueCounter;
+
+} ThreadSpecificData;
+/* static Tcl_ThreadDataKey dataKey; */
 
-static Tcl_Obj *whitespaceRE;
 /* This string is a backup.  Value should be defined in xml package. */
 static char whitespace[] = " \t\r\n";
 
-/*
- * Counter to generate unique command names
- */
-
-static int uniqueCounter = 0;
-
 /*
  * Configuration option tables
  */
@@ -57,6 +56,8 @@ static CONST84 char *instanceConfigureSwitches[] = {
     "-final",
     "-validate",
     "-baseurl", 
+    "-baseuri", 
+    "-encoding",
     "-elementstartcommand",
     "-elementendcommand",
     "-characterdatacommand",
@@ -86,31 +87,33 @@ static CONST84 char *instanceConfigureSwitches[] = {
     (char *) NULL
   };
 enum instanceConfigureSwitches {
-    TCLXML_FINAL, TCLXML_VALIDATE, TCLXML_BASE, 
-    TCLXML_ELEMENTSTARTCMD, TCLXML_ELEMENTENDCMD,
-    TCLXML_DATACMD, TCLXML_PICMD, 
-    TCLXML_DEFAULTCMD,
-    TCLXML_UNPARSEDENTITYCMD, TCLXML_NOTATIONCMD,
-    TCLXML_EXTERNALENTITYCMD, TCLXML_UNKNOWNENCODINGCMD,
-    TCLXML_COMMENTCMD, TCLXML_NOTSTANDALONECMD,
-    TCLXML_STARTCDATASECTIONCMD, TCLXML_ENDCDATASECTIONCMD,
-    TCLXML_DEFAULTEXPANDINTERNALENTITIES,
-    TCLXML_ELEMENTDECLCMD, TCLXML_ATTLISTDECLCMD,
-    TCLXML_STARTDOCTYPEDECLCMD, TCLXML_ENDDOCTYPEDECLCMD,
-    TCLXML_PARAMENTITYPARSING,
-    TCLXML_NOWHITESPACE,
-    TCLXML_REPORTEMPTY,
-    TCLXML_ENTITYDECLCMD,
-    TCLXML_PARAMENTITYDECLCMD,
-    TCLXML_DOCTYPECMD,
-    TCLXML_ENTITYREFCMD,
-    TCLXML_XMLDECLCMD
-  };
+  TCLXML_FINAL, TCLXML_VALIDATE, TCLXML_BASEURL, TCLXML_BASEURI, 
+  TCLXML_ENCODING,
+  TCLXML_ELEMENTSTARTCMD, TCLXML_ELEMENTENDCMD,
+  TCLXML_DATACMD, TCLXML_PICMD, 
+  TCLXML_DEFAULTCMD,
+  TCLXML_UNPARSEDENTITYCMD, TCLXML_NOTATIONCMD,
+  TCLXML_EXTERNALENTITYCMD, TCLXML_UNKNOWNENCODINGCMD,
+  TCLXML_COMMENTCMD, TCLXML_NOTSTANDALONECMD,
+  TCLXML_STARTCDATASECTIONCMD, TCLXML_ENDCDATASECTIONCMD,
+  TCLXML_DEFAULTEXPANDINTERNALENTITIES,
+  TCLXML_ELEMENTDECLCMD, TCLXML_ATTLISTDECLCMD,
+  TCLXML_STARTDOCTYPEDECLCMD, TCLXML_ENDDOCTYPEDECLCMD,
+  TCLXML_PARAMENTITYPARSING,
+  TCLXML_NOWHITESPACE,
+  TCLXML_REPORTEMPTY,
+  TCLXML_ENTITYDECLCMD,
+  TCLXML_PARAMENTITYDECLCMD,
+  TCLXML_DOCTYPECMD,
+  TCLXML_ENTITYREFCMD,
+  TCLXML_XMLDECLCMD
+};
 
 /*
  * Prototypes for procedures defined later in this file:
  */
 
+static void DeletePkgData _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp));
 static void TclXMLInstanceDeleteCmd _ANSI_ARGS_((ClientData clientData));
 static int  TclXMLDestroyParserInstance _ANSI_ARGS_((TclXML_Info *xmlinfo));
 static int  TclXMLInstanceCmd _ANSI_ARGS_((ClientData dummy,
@@ -186,7 +189,10 @@ int
 Tclxml_Init (interp)
       Tcl_Interp *interp; /* Interpreter to initialise. */
 {
-  extern TclxmlStubs tclxmlStubs;
+  /*
+  ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
+  */
+  ThreadSpecificData *tsdPtr;
 
 #ifdef USE_TCL_STUBS
   if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
@@ -194,25 +200,28 @@ Tclxml_Init (interp)
   }
 #endif
 
-  whitespaceRE = Tcl_GetVar2Ex(interp, "::xml::Wsp", NULL, TCL_GLOBAL_ONLY);
-  if (whitespaceRE == NULL) {
-    whitespaceRE = Tcl_SetVar2Ex(interp, "::xml::Wsp", NULL, Tcl_NewStringObj(whitespace, -1), TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
-    if (whitespaceRE == NULL) {
+  tsdPtr = (ThreadSpecificData *) Tcl_Alloc(sizeof(ThreadSpecificData));
+  Tcl_SetAssocData(interp, "::xml::c", DeletePkgData, (ClientData) tsdPtr);
+  tsdPtr->defaultParser = NULL;
+  tsdPtr->uniqueCounter = 0;
+
+  tsdPtr->whitespaceRE = Tcl_GetVar2Ex(interp, "::xml::Wsp", NULL, TCL_GLOBAL_ONLY);
+  if (tsdPtr->whitespaceRE == NULL) {
+    tsdPtr->whitespaceRE = Tcl_SetVar2Ex(interp, "::xml::Wsp", NULL, Tcl_NewStringObj(whitespace, -1), TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
+    if (tsdPtr->whitespaceRE == NULL) {
       return TCL_ERROR;
     }
   }
-  Tcl_IncrRefCount(whitespaceRE);
+  Tcl_IncrRefCount(tsdPtr->whitespaceRE);
 
-  xmlconfig = (TclXMLConfigInfo *) Tcl_Alloc(sizeof(TclXMLConfigInfo));
-  xmlconfig->defaultParser = NULL;
+  tsdPtr->registeredParsers = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
+  Tcl_InitHashTable(tsdPtr->registeredParsers, TCL_STRING_KEYS);
 
-  Tcl_InitHashTable(&registeredParsers, TCL_STRING_KEYS);
+  Tcl_CreateObjCommand(interp, "xml::configure", TclXMLConfigureCmd, (ClientData) tsdPtr, NULL);
+  Tcl_CreateObjCommand(interp, "xml::parser", TclXMLCreateParserCmd, (ClientData) tsdPtr, NULL);
+  Tcl_CreateObjCommand(interp, "xml::parserclass", TclXMLParserClassCmd, (ClientData) tsdPtr, NULL);
 
-  Tcl_CreateObjCommand(interp, "xml::configure", TclXMLConfigureCmd, NULL, NULL);
-  Tcl_CreateObjCommand(interp, "xml::parser", TclXMLCreateParserCmd, NULL, NULL);
-  Tcl_CreateObjCommand(interp, "xml::parserclass", TclXMLParserClassCmd, NULL, NULL);
-
-  #if TCL_DOES_STUBS
+  #if TCLXML_DOES_STUBS
     {
       extern TclxmlStubs tclxmlStubs;
       if (Tcl_PkgProvideEx(interp, "xml::c", TCLXML_VERSION,
@@ -228,7 +237,7 @@ Tclxml_Init (interp)
 
   return TCL_OK;
 }
-
+
 /*
  *----------------------------------------------------------------------------
  *
@@ -253,6 +262,36 @@ Tclxml_SafeInit (interp)
     return Tclxml_Init(interp);
 }
 
+/*
+ *----------------------------------------------------------------------------
+ *
+ * DeletePkgData --
+ *
+ *  Cleanup package memory usage.
+ *
+ * Results:
+ *  None.
+ *
+ * Side effects:
+ *  Memory deallocated.
+ *
+ *----------------------------------------------------------------------------
+ */
+
+static void
+DeletePkgData (clientData, interp)
+     ClientData clientData;
+     Tcl_Interp *interp;
+{
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) clientData;
+
+  Tcl_DecrRefCount(tsdPtr->whitespaceRE);
+  Tcl_DeleteHashTable(tsdPtr->registeredParsers);
+  Tcl_Free((char *) tsdPtr->registeredParsers);
+
+  Tcl_Free((char *) tsdPtr);
+}
+
 /*
  *----------------------------------------------------------------------------
  *
@@ -270,8 +309,8 @@ Tclxml_SafeInit (interp)
  */
 
 static int
-TclXMLConfigureCmd(dummy, interp, objc, objv)
-     ClientData dummy;
+TclXMLConfigureCmd(clientData, interp, objc, objv)
+     ClientData clientData;
      Tcl_Interp *interp;
      int objc;
      Tcl_Obj *CONST objv[];
@@ -297,12 +336,14 @@ TclXMLConfigureCmd(dummy, interp, objc, objv)
  */
 
 static int
-TclXMLParserClassCmd(dummy, interp, objc, objv)
-     ClientData dummy;
+TclXMLParserClassCmd(clientData, interp, objc, objv)
+     ClientData clientData;
      Tcl_Interp *interp;
      int objc;
      Tcl_Obj *CONST objv[];
 {
+  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) clientData;
   TclXML_ParserClassInfo *classinfo;
   int method, index;
   Tcl_Obj *listPtr;
@@ -445,9 +486,9 @@ TclXMLParserClassCmd(dummy, interp, objc, objv)
         case TCLXML_INFO_NAMES:
           
           listPtr = Tcl_NewListObj(0, NULL);
-          entryPtr = Tcl_FirstHashEntry(&registeredParsers, &search);
+          entryPtr = Tcl_FirstHashEntry(tsdPtr->registeredParsers, &search);
           while (entryPtr != NULL) {
-            Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(Tcl_GetHashKey(&registeredParsers, entryPtr), -1));
+            Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(Tcl_GetHashKey(tsdPtr->registeredParsers, entryPtr), -1));
             entryPtr = Tcl_NextHashEntry(&search);
           }
 
@@ -457,10 +498,10 @@ TclXMLParserClassCmd(dummy, interp, objc, objv)
 
         case TCLXML_INFO_DEFAULT:
 
-          if (!xmlconfig->defaultParser) {
+          if (!tsdPtr->defaultParser) {
             Tcl_SetResult(interp, "", NULL);
           } else {
-            Tcl_SetObjResult(interp, xmlconfig->defaultParser->name);
+            Tcl_SetObjResult(interp, tsdPtr->defaultParser->name);
           }
 
           break;
@@ -500,10 +541,14 @@ TclXML_RegisterXMLParser(interp, classinfo)
      Tcl_Interp *interp;
      TclXML_ParserClassInfo *classinfo;
 {
+  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
+  ThreadSpecificData *tsdPtr;
   int new;
   Tcl_HashEntry *entryPtr;
 
-  entryPtr = Tcl_CreateHashEntry(&registeredParsers, Tcl_GetStringFromObj(classinfo->name, NULL), &new);
+  tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
+
+  entryPtr = Tcl_CreateHashEntry(tsdPtr->registeredParsers, Tcl_GetStringFromObj(classinfo->name, NULL), &new);
   if (!new) {
     Tcl_Obj *ptr = Tcl_NewStringObj("parser class \"", -1);
     Tcl_AppendObjToObj(ptr, classinfo->name);
@@ -520,7 +565,7 @@ TclXML_RegisterXMLParser(interp, classinfo)
    * Set default parser - last wins
    */
 
-  xmlconfig->defaultParser = classinfo;
+  tsdPtr->defaultParser = classinfo;
 
   return TCL_OK;
 }
@@ -542,14 +587,17 @@ TclXML_RegisterXMLParser(interp, classinfo)
  */
 
 static int
-TclXMLCreateParserCmd(dummy, interp, objc, objv)
-     ClientData dummy;
+TclXMLCreateParserCmd(clientData, interp, objc, objv)
+     ClientData clientData;
      Tcl_Interp *interp;
      int objc;
      Tcl_Obj *CONST objv[];
 {
+  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
+  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
   TclXML_Info *xmlinfo;
   int found, i, index, poption;
+
   static CONST84 char *switches[] = {
     "-parser",
     (char *) NULL
@@ -558,8 +606,7 @@ TclXMLCreateParserCmd(dummy, interp, objc, objv)
     TCLXML_PARSER
   };
 
-
-  if (!xmlconfig->defaultParser) {
+  if (!tsdPtr->defaultParser) {
     Tcl_SetResult(interp, "no parsers available", NULL);
     return TCL_ERROR;
   }
@@ -575,6 +622,7 @@ TclXMLCreateParserCmd(dummy, interp, objc, objv)
   xmlinfo->interp = interp;
   xmlinfo->clientData = NULL;
   xmlinfo->base = NULL;
+  xmlinfo->encoding = Tcl_NewStringObj("utf-8", -1);
 
   /*
    * Find unique command name
@@ -686,7 +734,7 @@ TclXMLCreateParserCmd(dummy, interp, objc, objv)
      * Use given parser class
      */
 
-    pentry = Tcl_FindHashEntry(&registeredParsers,
+    pentry = Tcl_FindHashEntry(tsdPtr->registeredParsers,
 			       Tcl_GetStringFromObj(objv[poption + 1],
 						    NULL));
     if (pentry != NULL) {
@@ -702,7 +750,7 @@ TclXMLCreateParserCmd(dummy, interp, objc, objv)
     /*
      * Use default parser
      */
-    xmlinfo->parserClass = xmlconfig->defaultParser;
+    xmlinfo->parserClass = tsdPtr->defaultParser;
   }
 
   if (TclXMLResetParser(interp, xmlinfo) != TCL_OK) {
@@ -805,17 +853,19 @@ static Tcl_Obj *
 FindUniqueCmdName(interp)
      Tcl_Interp *interp;
 {
+  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
+  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
   Tcl_Obj *name;
-  Tcl_CmdInfo info;
+  Tcl_CmdInfo cmdinfo;
   char s[20];
 
   name = Tcl_NewStringObj("", 0);
   Tcl_IncrRefCount(name);
 
   do {
-    sprintf(s, "xmlparser%d", uniqueCounter++);
+    sprintf(s, "xmlparser%d", tsdPtr->uniqueCounter++);
     Tcl_SetStringObj(name, s, -1);
-  } while (Tcl_GetCommandInfo(interp, Tcl_GetStringFromObj(name, NULL), &info));
+  } while (Tcl_GetCommandInfo(interp, Tcl_GetStringFromObj(name, NULL), &cmdinfo));
 
   return name;
 }
@@ -1105,6 +1155,11 @@ TclXMLCreateEntityParser(interp, external, parent)
     external->context = NULL;
     external->cdata = NULL;
     external->nowhitespace = parent->nowhitespace;
+	if (parent->encoding) {
+	  external->encoding = Tcl_DuplicateObj(parent->encoding);
+	} else {
+	  external->encoding = Tcl_NewStringObj("utf-8", -1);
+	}
 
   if (extClassInfo->createEntity) {
     /*
@@ -1161,6 +1216,9 @@ TclXMLDestroyParserInstance(xmlinfo)
   if (xmlinfo->clientData) {
     if (classInfo->destroy) {
       if ((*classInfo->destroy)(xmlinfo->clientData) != TCL_OK) {
+		if (xmlinfo->encoding) {
+		  Tcl_DecrRefCount(xmlinfo->encoding);
+		}
         Tcl_Free((char *)xmlinfo);
         return TCL_ERROR;
       }
@@ -1177,6 +1235,9 @@ TclXMLDestroyParserInstance(xmlinfo)
       Tcl_Release((ClientData) xmlinfo->interp);
 
       if (result != TCL_OK) {
+		if (xmlinfo->encoding) {
+		  Tcl_DecrRefCount(xmlinfo->encoding);
+		}
         Tcl_Free((char *)xmlinfo);
         return TCL_ERROR;
       }
@@ -1213,9 +1274,12 @@ TclXMLFreeParser(xmlinfo)
      TclXML_Info *xmlinfo;
 {
 
-  TclXMLDestroyParserInstance(xmlinfo);
-
-  Tcl_Free((char*)xmlinfo);
+  if (TclXMLDestroyParserInstance(xmlinfo) == TCL_OK) {
+	if (xmlinfo->encoding) {
+	  Tcl_DecrRefCount(xmlinfo->encoding);
+	}
+	Tcl_Free((char*)xmlinfo);
+  }
 }
 
 /*
@@ -1243,8 +1307,9 @@ TclXMLInstanceCmd (clientData, interp, objc, objv)
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) clientData;
   TclXML_Info *child;
-  char *data;
+  char *encoding, *data;
   int len, index, result = TCL_OK;
+  Tcl_Obj *childNamePtr;
   static CONST84 char *options[] = {
     "configure", "cget", "entityparser", "free", "get", "parse", "reset", NULL
   };
@@ -1281,6 +1346,17 @@ TclXMLInstanceCmd (clientData, interp, objc, objv)
 
     case TCLXML_ENTITYPARSER:
       /* ericm@scriptics.com, 1999.9.13 */
+
+      /* check for args - Pat Thoyts */
+      if (objc == 2) {
+	childNamePtr = FindUniqueCmdName(interp);
+      } else if (objc == 3) {
+	childNamePtr = objv[2];
+      } else {
+        Tcl_WrongNumArgs(interp, 1, objv, "entityparser ?args?");
+        return TCL_ERROR;
+      }
+
       /*
        * Create the data structures for this parser.
        */
@@ -1290,19 +1366,14 @@ TclXMLInstanceCmd (clientData, interp, objc, objv)
         return TCL_ERROR;
       }
 
-      /* check for args - Pat Thoyts */
-      if (objc < 3) {
-        Tcl_WrongNumArgs(interp, 1, objv, "entityparser ?args?");
-        return TCL_ERROR;
-      }
-
       child->interp = interp;
-      Tcl_IncrRefCount(objv[2]);
-      child->name = objv[2];
+      Tcl_IncrRefCount(childNamePtr);
+      child->name = childNamePtr;
     
       /* Actually create the parser instance */
       if (TclXMLCreateEntityParser(interp, child,
         xmlinfo) != TCL_OK) {
+	Tcl_DecrRefCount(childNamePtr);
         Tcl_Free((char*)child);
         return TCL_ERROR;
       }
@@ -1335,7 +1406,17 @@ TclXMLInstanceCmd (clientData, interp, objc, objv)
         return TCL_ERROR;
       }
 
-      data = Tcl_GetStringFromObj(objv[2], &len);
+	  if (xmlinfo->encoding) {
+		encoding = Tcl_GetStringFromObj(xmlinfo->encoding, NULL);
+	  } else {
+		encoding = "utf-8";
+	  }
+	  if (strlen(encoding) == 0 || strcmp(encoding, "utf-8") == 0) {
+		data = Tcl_GetStringFromObj(objv[2], &len);
+	  } else {
+		data = Tcl_GetByteArrayFromObj(objv[2], &len);
+	  }
+
       result = TclXMLParse(interp, xmlinfo, data, len);
 
       break;
@@ -1425,6 +1506,7 @@ TclXMLParse (interp, xmlinfo, data, len)
     case TCL_OK:
     case TCL_BREAK:
     case TCL_CONTINUE:
+      TclXMLDispatchPCDATA(xmlinfo);
       Tcl_ResetResult(interp);
       return TCL_OK;
 
@@ -1465,61 +1547,66 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
-  int index, bool, doParse = 0;
-  Tcl_Obj *CONST *objPtr = objv;
+  int index, bool, doParse = 0, result;
   TclXML_ParserClassInfo *classinfo = (TclXML_ParserClassInfo *) xmlinfo->parserClass;
 
-  /*
-   * Firstly, pass the option to the parser's own 
-   * configuration management routine.
-   * It may pass back an error or break code to
-   * stop us from further processing the options.
-   */
-
-  if (classinfo->configure) {
-    if ((*classinfo->configure)(xmlinfo->clientData, objc, objv) != TCL_OK) {
-      return TCL_ERROR;
-    }
-  } else if (classinfo->configureCmd) {
-    Tcl_Obj *cmdPtr = Tcl_DuplicateObj(classinfo->configureCmd);
-    int i, result;
-
-    Tcl_IncrRefCount(cmdPtr);
-    Tcl_Preserve((ClientData) interp);
-
-    if (xmlinfo->clientData) {
-      Tcl_ListObjAppendElement(interp, cmdPtr, (Tcl_Obj *) xmlinfo->clientData);
-    } else if (xmlinfo->name) {
-      Tcl_ListObjAppendElement(interp, cmdPtr, xmlinfo->name);
-    }
-
-    for (i = 0; i < objc; i++) {
-      Tcl_ListObjAppendElement(interp, cmdPtr, objv[i]);
-    }
-
-    result = Tcl_GlobalEvalObj(interp, cmdPtr);
-
-    Tcl_DecrRefCount(cmdPtr);
-    Tcl_Release((ClientData) interp);
-
-    if (result == TCL_BREAK) {
-      return TCL_OK;
-    } else if (result != TCL_OK) {
-      return TCL_ERROR;
-    }
-  }
-
-  Tcl_ResetResult (interp);
-
   while (objc > 1) {
-    if (Tcl_GetIndexFromObj(interp, objPtr[0], instanceConfigureSwitches,
+    /*
+     * Firstly, pass the option to the parser's own 
+     * configuration management routine.
+     * It may pass back an error or break code to
+     * stop us from further processing the options.
+     */
+
+    if (classinfo->configure) {
+      result = (*classinfo->configure)(xmlinfo->clientData, objv[0], objv[1]);
+      if (result == TCL_BREAK) {
+	objc -= 2;
+	objv += 2;
+	continue;
+      }
+      if (result != TCL_OK) {
+	return TCL_ERROR;
+      }
+    } else if (classinfo->configureCmd) {
+      Tcl_Obj *cmdPtr = Tcl_DuplicateObj(classinfo->configureCmd);
+
+      Tcl_IncrRefCount(cmdPtr);
+      Tcl_Preserve((ClientData) interp);
+
+      if (xmlinfo->clientData) {
+	Tcl_ListObjAppendElement(interp, cmdPtr, (Tcl_Obj *) xmlinfo->clientData);
+      } else if (xmlinfo->name) {
+	Tcl_ListObjAppendElement(interp, cmdPtr, xmlinfo->name);
+      }
+
+      Tcl_ListObjAppendElement(interp, cmdPtr, objv[0]);
+      Tcl_ListObjAppendElement(interp, cmdPtr, objv[1]);
+
+      result = Tcl_GlobalEvalObj(interp, cmdPtr);
+
+      Tcl_DecrRefCount(cmdPtr);
+      Tcl_Release((ClientData) interp);
+
+      if (result == TCL_BREAK) {
+	objc -= 2;
+	objv += 2;
+	continue;
+      } else if (result != TCL_OK) {
+	return TCL_ERROR;
+      }
+    }
+
+    Tcl_ResetResult (interp);
+
+    if (Tcl_GetIndexFromObj(interp, objv[0], instanceConfigureSwitches,
                             "switch", 0, &index) != TCL_OK) {
       return TCL_ERROR;
     }
     switch ((enum instanceConfigureSwitches) index) {
       case TCLXML_FINAL:            /* -final */
 
-        if (Tcl_GetBooleanFromObj(interp, objPtr[1], &bool) != TCL_OK) {
+        if (Tcl_GetBooleanFromObj(interp, objv[1], &bool) != TCL_OK) {
           return TCL_ERROR;
         }
 
@@ -1537,8 +1624,16 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         xmlinfo->final = bool;
         break;
 
+	  case TCLXML_ENCODING: /* -encoding */
+		if (xmlinfo->encoding) {
+		  Tcl_DecrRefCount(xmlinfo->encoding);
+		}
+		xmlinfo->encoding = objv[1];
+		Tcl_IncrRefCount(xmlinfo->encoding);
+		break;
+
       case TCLXML_VALIDATE:         /* -validate */
-        if (Tcl_GetBooleanFromObj(interp, objPtr[1], &bool) != TCL_OK) {
+        if (Tcl_GetBooleanFromObj(interp, objv[1], &bool) != TCL_OK) {
           return TCL_ERROR;
         }
         /*
@@ -1548,18 +1643,19 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         xmlinfo->validate = bool;
         break;
 
-      case TCLXML_BASE:             /* -base */
+    case TCLXML_BASEURL:             /* -baseurl, -baseuri */
+    case TCLXML_BASEURI:
         if (xmlinfo->base != NULL) {
           Tcl_DecrRefCount(xmlinfo->base);
         }
 
-        xmlinfo->base = objPtr[1];
+        xmlinfo->base = objv[1];
         Tcl_IncrRefCount(xmlinfo->base);
         break;
 
       case TCLXML_DEFAULTEXPANDINTERNALENTITIES:    /* -defaultexpandinternalentities */
         /* ericm@scriptics */
-        if (Tcl_GetBooleanFromObj(interp, objPtr[1], &bool) != TCL_OK) {
+        if (Tcl_GetBooleanFromObj(interp, objv[1], &bool) != TCL_OK) {
           return TCL_ERROR;
         }
         xmlinfo->validate = bool;
@@ -1574,7 +1670,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
          * configure procedure.
          */
 
-        if (TclXMLConfigureParserInstance(xmlinfo, objPtr[0], objPtr[1]) != TCL_OK) {
+        if (TclXMLConfigureParserInstance(xmlinfo, objv[0], objv[1]) != TCL_OK) {
           return TCL_ERROR;
         }
         break;
@@ -1586,7 +1682,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->elementstart = NULL;
 
-        xmlinfo->elementstartcommand = objPtr[1];
+        xmlinfo->elementstartcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->elementstartcommand);
         break;
 
@@ -1597,7 +1693,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->elementend = NULL;
 
-        xmlinfo->elementendcommand = objPtr[1];
+        xmlinfo->elementendcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->elementendcommand);
         break;
 
@@ -1608,7 +1704,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->cdatacb = NULL;
 
-        xmlinfo->datacommand = objPtr[1];
+        xmlinfo->datacommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->datacommand);
         break;
 
@@ -1619,7 +1715,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->pi = NULL;
 
-        xmlinfo->picommand = objPtr[1];
+        xmlinfo->picommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->picommand);
         break;
 
@@ -1630,7 +1726,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->defaultcb = NULL;
 
-        xmlinfo->defaultcommand = objPtr[1];
+        xmlinfo->defaultcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->defaultcommand);
         break;
 
@@ -1641,7 +1737,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->unparsed = NULL;
 
-        xmlinfo->unparsedcommand = objPtr[1];
+        xmlinfo->unparsedcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->unparsedcommand);
         break;
 
@@ -1652,7 +1748,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->notation = NULL;
 
-        xmlinfo->notationcommand = objPtr[1];
+        xmlinfo->notationcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->notationcommand);
         break;
 
@@ -1663,7 +1759,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->entity = NULL;
 
-        xmlinfo->entitycommand = objPtr[1];
+        xmlinfo->entitycommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->entitycommand);
         break;
 
@@ -1677,7 +1773,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->unknownencoding = NULL;
 
-        xmlinfo->unknownencodingcommand = objPtr[1];
+        xmlinfo->unknownencodingcommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->unknownencodingcommand);
         break;
     
@@ -1688,7 +1784,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->comment = NULL;
 
-        xmlinfo->commentCommand = objPtr[1];
+        xmlinfo->commentCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->commentCommand);
         break;
 
@@ -1699,7 +1795,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->notStandalone = NULL;
 
-        xmlinfo->notStandaloneCommand = objPtr[1];
+        xmlinfo->notStandaloneCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->notStandaloneCommand);
         break;
 
@@ -1711,7 +1807,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->startCDATASection = NULL;
 
-        xmlinfo->startCdataSectionCommand = objPtr[1];
+        xmlinfo->startCdataSectionCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->startCdataSectionCommand);
         break;
 
@@ -1722,7 +1818,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->endCDATASection = NULL;
 
-        xmlinfo->endCdataSectionCommand = objPtr[1];
+        xmlinfo->endCdataSectionCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->endCdataSectionCommand);
         break;
 #endif
@@ -1734,7 +1830,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->elementDecl = NULL;
 
-        xmlinfo->elementDeclCommand = objPtr[1];
+        xmlinfo->elementDeclCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->elementDeclCommand);
         break;
 
@@ -1745,7 +1841,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->attlistDecl = NULL;
 
-        xmlinfo->attlistDeclCommand = objPtr[1];
+        xmlinfo->attlistDeclCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->attlistDeclCommand);
         break;
 
@@ -1756,7 +1852,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->startDoctypeDecl = NULL;
 
-        xmlinfo->startDoctypeDeclCommand = objPtr[1];
+        xmlinfo->startDoctypeDeclCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->startDoctypeDeclCommand);
         break;
 
@@ -1767,7 +1863,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         }
         xmlinfo->endDoctypeDecl = NULL;
 
-        xmlinfo->endDoctypeDeclCommand = objPtr[1];
+        xmlinfo->endDoctypeDeclCommand = objv[1];
         Tcl_IncrRefCount(xmlinfo->endDoctypeDeclCommand);
         break;
 
@@ -1784,7 +1880,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         break;
     }
 
-    objPtr += 2;
+    objv += 2;
     objc -= 2;
 
   }
@@ -1847,11 +1943,17 @@ TclXMLCget (interp, xmlinfo, objc, objv)
     case TCLXML_NOWHITESPACE:
       Tcl_SetObjResult(interp, Tcl_NewBooleanObj(xmlinfo->nowhitespace));
       break;
-    case TCLXML_BASE:
+    case TCLXML_BASEURL:
+	case TCLXML_BASEURI:
       if (xmlinfo->base) {
         Tcl_SetObjResult(interp, xmlinfo->base);
       }
       break;
+	case TCLXML_ENCODING:
+	  if (xmlinfo->encoding) {
+		Tcl_SetObjResult(interp, xmlinfo->encoding);
+	  }
+	  break;
     case TCLXML_ELEMENTSTARTCMD:
       if (xmlinfo->elementstartcommand) {
         Tcl_SetObjResult(interp, xmlinfo->elementstartcommand);
@@ -1918,6 +2020,10 @@ TclXMLCget (interp, xmlinfo, objc, objv)
         Tcl_SetObjResult(interp, xmlinfo->endCdataSectionCommand);
       }
       break;
+#else
+    case TCLXML_STARTCDATASECTIONCMD:
+    case TCLXML_ENDCDATASECTIONCMD:
+      break;
 #endif
     case TCLXML_ELEMENTDECLCMD:
       if (xmlinfo->elementDeclCommand) {
@@ -1939,6 +2045,14 @@ TclXMLCget (interp, xmlinfo, objc, objv)
         Tcl_SetObjResult(interp, xmlinfo->endDoctypeDeclCommand);
       }
       break;
+
+  case TCLXML_ENTITYDECLCMD:
+  case TCLXML_PARAMENTITYDECLCMD:
+  case TCLXML_DOCTYPECMD:
+  case TCLXML_ENTITYREFCMD:
+  case TCLXML_XMLDECLCMD:
+    /* These are not (yet) supported) */
+    break;
   }
 
   return TCL_OK;
@@ -1969,16 +2083,9 @@ TclXMLConfigureParserInstance (xmlinfo, option, value)
   TclXML_ParserClassInfo *classInfo = (TclXML_ParserClassInfo *) xmlinfo->parserClass;
 
   if (classInfo->configure) {
-    Tcl_Obj *objv[3];
-
-    objv[0] = option;
-    objv[1] = value;
-    objv[2] = NULL;
-
-    if ((*classInfo->configure)(xmlinfo->clientData, 2, objv) != TCL_OK) {
+    if ((*classInfo->configure)(xmlinfo->clientData, option, value) != TCL_OK) {
       return TCL_ERROR;
     }
-
   } else if (classInfo->configureCmd) {
     Tcl_Obj *cmdPtr = Tcl_DuplicateObj(classInfo->configureCmd);
     int result;
@@ -2098,7 +2205,7 @@ TclXMLHandlerResult(xmlinfo, result)
        * elements.
        */
       xmlinfo->status = TCL_CONTINUE;
-      xmlinfo->continueCount = 1;
+      xmlinfo->continueCount = 0;
       break;
 
     case TCL_BREAK:
@@ -2137,13 +2244,15 @@ TclXMLHandlerResult(xmlinfo, result)
  */
 
 void
-TclXML_ElementStartHandler(userData, name, atts)
+TclXML_ElementStartHandler(userData, name, nsuri, atts, nsDecls)
      void *userData;
      Tcl_Obj *name;
+     Tcl_Obj *nsuri;
      Tcl_Obj *atts;
+     Tcl_Obj *nsDecls;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2165,7 +2274,7 @@ TclXML_ElementStartHandler(userData, name, atts)
   }
 
   if (xmlinfo->elementstart) {
-    result = (xmlinfo->elementstart)(xmlinfo->interp, xmlinfo->elementstartdata, name, atts);
+    result = (xmlinfo->elementstart)(xmlinfo->interp, xmlinfo->elementstartdata, name, nsuri, atts, nsDecls);
   } else if (xmlinfo->elementstartcommand) {
     Tcl_Obj *cmdPtr;
 
@@ -2179,6 +2288,17 @@ TclXML_ElementStartHandler(userData, name, atts)
 
     Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, name);
     Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, atts);
+    if (nsuri) {
+      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewStringObj("-namespace", -1));
+      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, nsuri);
+    }
+    if (nsDecls) {
+      int len;
+      if ((Tcl_ListObjLength(xmlinfo->interp, nsDecls, &len) == TCL_OK) && (len > 0)) {
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewStringObj("-namespacedecls", -1));
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, nsDecls);
+      }
+    }
 
     /*
      * It would be desirable to be able to terminate parsing
@@ -2216,7 +2336,7 @@ TclXML_ElementEndHandler(userData, name)
      Tcl_Obj *name;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2228,8 +2348,9 @@ TclXML_ElementEndHandler(userData, name)
 
     if (!--(xmlinfo->continueCount)) {
       xmlinfo->status = TCL_OK;
+    } else {
       return;
-    }
+	}
   }
 
   if ((xmlinfo->elementend == NULL && 
@@ -2297,7 +2418,6 @@ TclXML_CharacterDataHandler(userData, s)
     Tcl_AppendObjToObj(xmlinfo->cdata, s);
   }
 }
-
 
 /*
  *----------------------------------------------------------------------------
@@ -2320,7 +2440,9 @@ static void
 TclXMLDispatchPCDATA(xmlinfo)
      TclXML_Info *xmlinfo;
 {
-  int result;
+  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
+  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(xmlinfo->interp, "::xml::c", NULL);
+  int result = TCL_OK;
 
   if (xmlinfo->cdata == NULL || 
       (xmlinfo->datacommand == NULL && xmlinfo->cdatacb == NULL) ||
@@ -2333,7 +2455,7 @@ TclXMLDispatchPCDATA(xmlinfo)
    */
 
   if (xmlinfo->nowhitespace) {
-    if (!Tcl_RegExpMatchObj(xmlinfo->interp, xmlinfo->cdata, whitespaceRE)) {
+    if (!Tcl_RegExpMatchObj(xmlinfo->interp, xmlinfo->cdata, tsdPtr->whitespaceRE)) {
       goto finish;
     }
   }
@@ -2398,7 +2520,7 @@ TclXML_ProcessingInstructionHandler(userData, target, data)
      Tcl_Obj *data;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2460,7 +2582,7 @@ TclXML_DefaultHandler(userData, s)
      Tcl_Obj *s;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2525,7 +2647,7 @@ TclXML_UnparsedDeclHandler(userData, entityName, base, systemId, publicId, notat
      Tcl_Obj *notationName;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2601,7 +2723,7 @@ TclXML_NotationDeclHandler(userData, notationName, base, systemId, publicId)
      Tcl_Obj *publicId;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2675,7 +2797,7 @@ TclXML_UnknownEncodingHandler(encodingHandlerData, name, info)
      void *info;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) encodingHandlerData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2745,9 +2867,9 @@ TclXML_ExternalEntityRefHandler(userData, openEntityNames, base,
      Tcl_Obj *publicId;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
   Tcl_Obj *oldContext;
-  
+
   TclXMLDispatchPCDATA(xmlinfo);
 
   if ((xmlinfo->entitycommand == NULL && xmlinfo->entity == NULL) || 
@@ -2824,7 +2946,7 @@ TclXML_CommentHandler(userData, data)
     Tcl_Obj *data;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2887,7 +3009,7 @@ TclXML_NotStandaloneHandler(userData)
     void *userData;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2940,7 +3062,7 @@ TclXML_ElementDeclHandler(userData, name, contentspec)
     Tcl_Obj *contentspec;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -2996,7 +3118,7 @@ TclXML_AttlistDeclHandler(userData, name, attributes)
     Tcl_Obj *attributes;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -3051,7 +3173,7 @@ TclXML_StartDoctypeDeclHandler(userData, name)
     Tcl_Obj *name;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
@@ -3103,7 +3225,7 @@ TclXML_EndDoctypeDeclHandler(userData)
     void *userData;
 {
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
-  int result;
+  int result = TCL_OK;
 
   TclXMLDispatchPCDATA(xmlinfo);
 
